/**
# Programa para calibrar una cámara a partir
# de una sola imagen (se estima solo la distancia focal),
# usando un patrón de dos círculos concéntricos
# Fraga 28/06/2011
  Versión en C: 
  Fraga 18/07/2017
**/
#include <stdio.h>
#include <math.h>

extern void dsyevv3( double A[][3], double V[][3], double *e);

void imprime_matriz( double a[][3], char *s )
{
    int i, j;


    printf ("--- %s----\n", s );
    for (i=0; i<3; i++){
        for (j=0; j<3; j++)
            printf ("%g ", a[i][j]);
        printf ("\n");
    }
}

double cubic_root( double v ) 
{
	double signo = 1.0;
	double del;

	if( v < 0.0 ) {
		signo = -1.0;
		v *= -1.0;
	}
	del = signo*exp( log(v)/3.0 );

	return( del );
}

/** 
 Args: ax^3 + bx^2 + cx + d = 0
 double * roots must be pointer to an array of size three 
 Returns the number of real root and their values insise roots[]
**/

int cubic( double a, double b, double c, double d, double *roots )
{
	double xN, yN, yN2;
	double delta2, h2, delta;

	xN = -b/(3.0*a);

	yN = xN*( xN*(a*xN + b) + c ) + d;
	yN2 = yN*yN;

	delta2 = (b*b - 3.0*a*c)/(9.0*a*a); 
	h2 = 4.0*a*a*delta2*delta2*delta2;

	if ( fabs(yN2 - h2) < 1e-12 ) {
		if ( fabs( yN2 ) < 1e-12 ){
			// Three equal roots in xN 
			roots[0] = roots[1] = roots[2] = xN;
			return( 3 );
		}
		else {
			// printf ("Dos reales iguales y la tercera distinta\n" );
			delta = cubic_root( yN/(2.0*a) );
			roots[0] = xN + delta;
			roots[1] = roots[2] = xN - 2.0*delta;
			return( 3 );
		}
	}
	else if ( yN2 > h2 ) {
		double v, v1, v2;

		// printf( "Una raiz real\n" );
		v = sqrt( yN2 - h2 );
		v1 = cubic_root( (-yN + v)/(2.0*a) );
		v2 = cubic_root( (-yN - v)/(2.0*a) );

		roots[0] = xN + v1 + v2;
		// print "Raiz en $r1\n";
		return( 1 );
	}
	else {
		double c, s, theta;

		// print "Tres raices reales distintas\n";

		c = -yN/sqrt( h2 );
		s = sqrt( 1.0 - c*c );
		theta = atan2( s, c )/3.0;
		// print "Theta = $theta\n";
		delta = 2.0*sqrt( delta2 );
	
		roots[0] = xN + delta*cos( theta );
		roots[1] = xN + delta*cos( theta + 2.09439510266 );
		roots[2] = xN + delta*cos( theta + 4.18879020533 );
		// print "Raices: $r1 $r2 $r3\n";

		return( 3 );
	}
}

/**
   Calcula los eigenvalores de una matriz de tamaño 3x3
**/
int eig( double A[][3], double *e )
{
	double b, c, d;
/*
                                     2                    2              2
(%o5) - s00 s11 s22 + l s11 s22 + s01  s22 + l s00 s22 - l  s22 + s00 s12
        2                      2                    2            2        2
 - l s12  - 2 s01 s02 s12 + s02  s11 + l s00 s11 - l  s11 - l s02  - l s01
    2        3
 - l  s00 + l  = 0
*/

	b = -A[0][0] - A[1][1] - A[2][2];
	c = A[1][1]*A[2][2] + A[0][0]*A[2][2] - A[1][2]*A[1][2] + A[0][0]*A[1][1] - A[0][2]*A[0][2] - A[0][1]*A[0][1];
	d = A[0][1]*A[0][1]*A[2][2] - A[0][0]*A[1][1]*A[2][2] + A[0][0]*A[1][2]*A[1][2] - 2.0*A[0][1]*A[0][2]*A[1][2] + A[0][2]*A[0][2]*A[1][1];

	if( cubic( 1.0, b, c, d, e ) == 3 ) {
		return 3;
	}
	return -1;
}

/*
  Eigenvalores de A, donde A no es simétrica
*/
int eigNonSym( double A[][3], double *e )
{
	double b, c, d;
/**
                                                             2
(%o5) - A00 A11 A22 + l A11 A22 + A01 A10 A22 + l A00 A22 - l  A22
 + A00 A12 A21 - l A12 A21 - A02 A10 A21 - A01 A12 A20 + A02 A11 A20
                            2                    2        3
 - l A02 A20 + l A00 A11 - l  A11 - l A01 A10 - l  A00 + l  = 0
**/

	b = -A[0][0] - A[1][1] - A[2][2];
	c = A[1][1]*A[2][2] + A[0][0]*A[2][2] - A[1][2]*A[2][1] + A[0][0]*A[1][1] - A[0][2]*A[2][0] - A[0][1]*A[1][0];
	d = A[0][1]*A[1][0]*A[2][2] - A[0][0]*A[1][1]*A[2][2] + A[0][0]*A[1][2]*A[2][1] - A[0][2]*A[1][0]*A[2][1] - A[0][1]*A[1][2]*A[2][0] + A[0][2]*A[1][1]*A[2][0];

	if( cubic( 1.0, b, c, d, e ) == 3 ) {
		return 3;
	}
	return -1;
}

/* 
   Realiza C = A * B
*/
void multiplicaAB(double A[][3], double B[][3], double C[][3])
{
	int i, j;
	for (i=0; i<3;i++)
		for (j=0;j<3;j++)
			C[i][j] = A[i][0]*B[0][j] + A[i][1]*B[1][j] + A[i][2]*B[2][j];
}

/* 
   Realiza C = A * B^T
*/
void multiplicaABT(double A[][3], double B[][3], double C[][3])
{
	int i, j;
	for (i=0; i<3;i++)
		for (j=0;j<3;j++)
			C[i][j] = A[i][0]*B[j][0] + A[i][1]*B[j][1] + A[i][2]*B[j][2];
}


// Ai = V D^{-1] V^T 
void inv( double A[][3], double Ai[][3] )
{
	int i, j;
	double V[3][3], e[3], T[3][3];

	dsyevv3( A, V, e);
	// imprime_matriz( V, "v.. ");
	// printf("%g %g %g\n", e[0], e[1], e[2] );

    for( i=0; i<3; i++ )
        for( j=0; j<3; j++ ) Ai[i][j] = 0.0;

    Ai[0][0] = 1.0/e[0];
    Ai[1][1] = 1.0/e[1];
    Ai[2][2] = 1.0/e[2];

    //  T = Eigvec * M
    multiplicaAB( V, Ai, T );
    multiplicaABT( T, V, Ai );  
}

void ordena( double *e )
{
	double tmp;
	
	if( e[0] > e[1] ) {
		tmp = e[0];
		e[0] = e[1];
		e[1] = tmp;
	}
	if( e[1] > e[2] ) {
		tmp = e[1];
		e[1] = e[2];
		e[2] = tmp;
	}
	if( e[0] > e[1] ) {
		tmp = e[0];
		e[0] = e[1];
		e[1] = tmp;
	}
}

void copia( double A[][3], double B[][3] )
{
	int i, j;

	for( i=0; i<3; i++ ) 
		for( j=0; j<3; j++ ) B[i][j] = A[i][j];
}


// Aqui se normaliza la elipse según A(3,3)
void normaliza( double A[][3] )
{
	int i, j;
	double v;

	v = A[2][2]*0.001;
	for( i=0; i<3; i++ ) 
		for( j=0; j<3; j++ ) A[i][j] /= v;
}


int main( )
{
	// Las matrices de los dos círculos proyectados
	double A1[3][3] = { 1.0000e-06,  -1.5031e-14,  -3.0000e-04,
	  -1.5031e-14,   6.6631e-06,  -1.1775e-03,
	  -3.0000e-04,  -1.1775e-03,   2.3139e-01 };

	double A2[3][3] = { 1.0000e-06,  -1.5525e-14,  -3.0000e-04,
  	-1.5525e-14,   6.9491e-06,  -1.3510e-03,
  	-3.0000e-04,  -1.3510e-03,   3.3668e-01 };

	double A1i[3][3], A2i[3][3];
	double e[3], ev[3];
	double C[3][3], val, arreglo[3];
	double E[3][3];
	double J[3][3];
	double u0, v0, f;
	double Ki[3][3];
	int i, j;

	normaliza( A1 );
	copia(A1, C );
	inv( C, A1i );

	normaliza( A2 );
	copia(A2, C );
	inv( C, A2i );

	multiplicaAB( A1, A2i, C );
	if( eigNonSym( C, e ) < 0 ) 
		printf("Esto no debería pasar\n");


	// Se ordenan los eigenvalores
	ordena( e );
	// printf( "%lf %lf %lf\n", e[0], e[1], e[2] );

	for( i=0; i<3; i++ ) 
		arreglo[i] = fabs( e[1]-e[i] );

	i = 0;
	if ( arreglo[1] > arreglo[i] ) i = 1;
	if ( arreglo[2] > arreglo[i] ) i = 2;
	val = e[i];
	// printf("Raiz unica: %lf\n", val);

	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ )
			C[i][j] = val*A1i[i][j] - A2i[i][j];

	dsyevv3( C, E, ev );
	
	ev[1] = sqrt( ev[1] ); 
	ev[2] = sqrt( ev[2] ); 

	for( i=0; i<3; i++ ) {
		J[i][0] = ev[1]*E[i][1];	
		J[i][1] = ev[2]*E[i][2];	
	}


	//  La raiz repetida
	i = 0;
	if ( arreglo[1] < arreglo[i] ) i = 1;
	if ( arreglo[2] < arreglo[i] ) i = 2;
	val = e[i];

	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ )
			C[i][j] = val*A1i[i][j] - A2i[i][j];

	dsyevv3( C, E, ev );
	printf("EV: %lf %lf %lf\n", ev[0], ev[1], ev[2] );

	val = sqrt( ev[2] ); 

	for( i=0; i<3; i++ ) 
		J[i][2] = val*E[i][2];	

	// imprime_matriz( J, "J " );
	u0 = 300.0;
	v0 = 200.0;
	f = - u0*u0 - v0*v0 + u0*(J[0][1]/J[2][1] + J[0][0]/J[2][0]) + 
	         v0*(J[1][1]/J[2][1] + J[1][0]/J[2][0]) 
	       - J[1][0]*J[1][1]/(J[2][0]*J[2][1]) - J[0][0]*J[0][1]/(J[2][0]*J[2][1]);

	f = sqrt(f);
	printf("f: %lf\n", f );

	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ )
			Ki[i][j] = 0.0;

	Ki[0][0] = Ki[1][1] =  1.0/f;
	Ki[0][2] = -u0/f;
	Ki[1][2] = -v0/f;
	Ki[2][2] = -1.0;

	multiplicaAB( Ki, J, C );

	val = C[0][0]*C[0][0] + C[1][0]*C[1][0] + C[2][0]*C[2][0];
	val = sqrt(val);
	for( i=0; i<3; i++ )
		for( j=0; j<3; j++ )
			C[i][j] /= val;

	imprime_matriz( C, "r1 r2 t " );

	/**
R = M1;

%% Corrección para escoger la orientación correcta:
v = cross ( R(:,1), R(:,2) )
v2 = [ 0 0 1] * v;
if v2 < 0  
    R(:,1) = -R(:,1)
    R( :,3 ) = cross( R( :,1 ), R( :,2 ) );
end
	**/

	return 0;
}
